@@ -1,22 +1,17 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals, print_function
+import os, sys, io, re, unittest, json, enum, textwrap, collections as cs
 
-import itertools as it, operator as op, functools as ft
-import os, sys, io, yaml, unittest
-
-if sys.version_info.major > 2:
-	from collections import OrderedDict, namedtuple
-	from collections.abc import Mapping
-	unicode = str
-else: from collections import Mapping, OrderedDict, namedtuple
+import yaml
 
 try: import pyaml
 except ImportError:
 	sys.path.insert(1, os.path.join(__file__, *['..']*3))
 	import pyaml
 
+try: import unidecode
+except ImportError: unidecode = None
 
-large_yaml = b'''
+
+large_yaml = br'''
 ### Default (baseline) configuration parameters.
 ### DO NOT ever change this config, use -c commandline option instead!
 
@@ -160,20 +155,24 @@ logging: # see http://docs.python.org/library/logging.config.html
     handlers: [console]
 '''
 
+class test_const(enum.IntEnum):
+	dispatch = 2455
+	heartbeat = 123
+
 data = dict(
 	path='/some/path',
-	query_dump=OrderedDict([
+	query_dump=cs.OrderedDict([
 		('key1', 'тест1'),
 		('key2', 'тест2'),
 		('key3', 'тест3'),
 		('последний', None) ]),
-	ids=OrderedDict(),
-	a=[1,None,'asd', 'не-ascii'], b=3.5, c=None,
-	asd=OrderedDict([('b', 1), ('a', 2)]) )
+	ids=cs.OrderedDict(),
+	a=[1,None,'asd', 'не-ascii'], b=3.5, c=None, d=test_const.dispatch,
+	asd=cs.OrderedDict([('b', 1), ('a', 2)]) )
 data['query_dump_clone'] = data['query_dump']
 data['ids']['id в уникоде'] = [4, 5, 6]
 data['ids']['id2 в уникоде'] = data['ids']['id в уникоде']
-# data["'asd'\n!\0\1"] =OrderedDict([('b', 1), ('a', 2)]) <-- fails in many ways
+data["'asd'\n!\0\1"] = cs.OrderedDict([('b', 1), ('a', 2)])
 
 data_str_multiline = dict(cert=(
 	'-----BEGIN CERTIFICATE-----\n'
@@ -218,110 +217,102 @@ data_str_long = dict(cert=(
 	'2sn5m3EkrUb1hSKQbMWTbnaG4C/F3i4KVkH+8AZmR9OvOmZ+7Lo=' ))
 
 
-# Restore Python2-like heterogeneous list sorting functionality in Python3
-# Based on https://gist.github.com/pR0Ps/1e1a1e892aad5b691448
-def compare(x, y):
-	if x == y: return 0
-	try:
-		if x < y: return -1
-		else: return 1
-
-	except TypeError as e:
-		# The case where both are None is taken care of by the equality test
-		if x is None: return -1
-		elif y is None: return 1
-
-		if type(x) != type(y):
-			return compare(*map(lambda t: type(t).__name__, [x, y]))
-
-		# Types are the same but a native compare didn't work.
-		# x and y might be indexable, recursively compare elements
-		for a, b in zip(x, y):
-			c = compare(a, b)
-			if c != 0: return c
-
-		return compare(len(x), len(y))
-
-
 class DumpTests(unittest.TestCase):
 
+	def yaml_var(self, ys, raw=False):
+		ys = textwrap.dedent(ys).replace('\t', '  ')
+		return ys if raw else yaml.safe_load(ys)
+
 	def flatten(self, data, path=tuple()):
 		dst = list()
 		if isinstance(data, (tuple, list)):
-			for v in data:
-				dst.extend(self.flatten(v, path + (list,)))
-		elif isinstance(data, Mapping):
-			for k,v in data.items():
-				dst.extend(self.flatten(v, path + (k,)))
+			for v in data: dst.extend(self.flatten(v, path + ('!!list',)))
+		elif isinstance(data, dict):
+			for k,v in data.items(): dst.extend(self.flatten(v, path + (k,)))
 		else: dst.append((path, data))
-		return tuple(sorted(dst, key=ft.cmp_to_key(compare)))
+		return tuple(sorted(dst, key=lambda v: json.dumps(v, sort_keys=True)))
+
+	def pos_list(self, ys, sep='\n'):
+		pos, pos_list = 0, list()
+		while True:
+			pos = ys.find(sep, pos+1)
+			if pos < 0: break
+			pos_list.append(pos)
+		return pos_list
+
+	def empty_line_list(self, ys):
+		return list(n for n, line in enumerate(ys.splitlines()) if not line)
+
 
 	def test_dst(self):
 		buff = io.BytesIO()
 		self.assertIs(pyaml.dump(data, buff), None)
 		self.assertIsInstance(pyaml.dump(data, str), str)
-		self.assertIsInstance(pyaml.dump(data, unicode), unicode)
+		self.assertIsInstance(pyaml.dump(data, bytes), bytes)
 
 	def test_simple(self):
 		a = self.flatten(data)
-		b = pyaml.dump(data, unicode)
+		b = pyaml.dump(data)
 		self.assertEqual(a, self.flatten(yaml.safe_load(b)))
 
 	def test_vspacing(self):
 		data = yaml.safe_load(large_yaml)
 		a = self.flatten(data)
-		b = pyaml.dump(data, unicode, vspacing=[2, 1])
+		b = pyaml.dump(data, vspacing=dict(split_lines=10, split_count=2))
 		self.assertEqual(a, self.flatten(yaml.safe_load(b)))
-		pos, pos_list = 0, list()
-		while True:
-			pos = b.find(u'\n', pos+1)
-			if pos < 0: break
-			pos_list.append(pos)
-		self.assertEqual( pos_list,
-			[ 12, 13, 25, 33, 53, 74, 89, 108, 158, 185, 265, 300, 345, 346, 356, 376, 400, 426, 427,
-				460, 461, 462, 470, 508, 564, 603, 604, 605, 611, 612, 665, 666, 690, 691, 715, 748,
-				777, 806, 807, 808, 817, 818, 832, 843, 878, 948, 949, 961, 974, 1009, 1032, 1052,
-				1083, 1102, 1123, 1173, 1195, 1234, 1257, 1276, 1300, 1301, 1312, 1325, 1341, 1359,
-				1374, 1375, 1383, 1397, 1413, 1431, 1432, 1453, 1454, 1467, 1468, 1485, 1486, 1487,
-				1498, 1499, 1530, 1531, 1551, 1552, 1566, 1577, 1590, 1591, 1612, 1613, 1614, 1622,
-				1623, 1638, 1648, 1649, 1657, 1658, 1688, 1689, 1698, 1720, 1730 ] )
-		b = pyaml.dump(data, unicode)
+		self.assertEqual( self.pos_list(b, '\n'),
+			[12, 13, 25, 33, 52, 73, 88, 107, 157, 184, 264, 299, 344, 345, 355, 375, 399,
+			424, 425, 458, 459, 467, 505, 561, 600, 601, 607, 660, 681, 705, 738, 767, 795,
+			796, 805, 806, 820, 831, 866, 936, 937, 949, 950, 963, 998, 1021, 1041, 1072,
+			1073, 1092, 1113, 1163, 1185, 1224, 1247, 1266, 1290, 1291, 1302, 1315, 1331,
+			1349, 1364, 1365, 1373, 1387, 1403, 1421, 1422, 1440, 1441, 1454, 1455, 1471,
+			1472, 1483, 1511, 1528, 1542, 1553, 1566, 1584, 1585, 1593, 1608, 1618, 1626,
+			1656, 1665, 1686, 1696] )
+		b = pyaml.dump(data, vspacing=False)
 		self.assertNotIn('\n\n', b)
 
 	def test_ids(self):
-		b = pyaml.dump(data, unicode)
+		b = pyaml.dump(data, force_embed=False)
 		self.assertNotIn('&id00', b)
 		self.assertIn('query_dump_clone: *query_dump_clone', b)
-		self.assertIn("'id в уникоде': &ids_-_id2_v_unikode", b) # kinda bug - should be just "id"
+		self.assertIn('id в уникоде: &ids_-_id2_', b)
+		if not unidecode: self.assertIn('id в уникоде: &ids_-_id2__id00', b)
+
+	def test_ids_unidecode(self):
+		if not unidecode:
+			self.skipTest('No unidecode module to test ids from non-ascii keys')
+		b = pyaml.dump(data, force_embed=False)
+		self.assertNotIn('&id00', b)
+		self.assertNotIn('_id00', b)
+		self.assertIn('id в уникоде: &ids_-_id2_v_unikode', b)
 
 	def test_force_embed(self):
-		b = pyaml.dump(data, unicode, force_embed=True)
-		c = pyaml.dump(data, unicode, safe=True, force_embed=True)
-		for char, dump in it.product('*&', [b, c]):
-			self.assertNotIn(char, dump)
+		for check, fe in (self.assertNotIn, True), (self.assertIn, False):
+			dump = pyaml.dump(data, force_embed=fe)
+			for c in '*&': check(c, dump)
 
 	def test_encoding(self):
-		b = pyaml.dump(data, unicode, force_embed=True)
-		b_lines = list(map(unicode.strip, b.splitlines()))
+		b = pyaml.dump(data, force_embed=True)
+		b_lines = list(map(str.strip, b.splitlines()))
 		chk = ['query_dump:', 'key1: тест1', 'key2: тест2', 'key3: тест3', 'последний:']
 		pos = b_lines.index('query_dump:')
 		self.assertEqual(b_lines[pos:pos + len(chk)], chk)
 
 	def test_str_long(self):
-		b = pyaml.dump(data_str_long, unicode)
+		b = pyaml.dump(data_str_long)
 		self.assertNotIn('"', b)
 		self.assertNotIn("'", b)
 		self.assertEqual(len(b.splitlines()), 1)
 
 	def test_str_multiline(self):
-		b = pyaml.dump(data_str_multiline, unicode)
+		b = pyaml.dump(data_str_multiline)
 		b_lines = b.splitlines()
 		self.assertGreater(len(b_lines), len(data_str_multiline['cert'].splitlines()))
 		for line in b_lines: self.assertLess(len(line), 100)
 
 	def test_dumps(self):
 		b = pyaml.dumps(data_str_multiline)
-		self.assertIsInstance(b, bytes)
+		self.assertIsInstance(b, str)
 
 	def test_print(self):
 		self.assertIs(pyaml.print, pyaml.pprint)
@@ -338,14 +329,28 @@ class DumpTests(unittest.TestCase):
 		pyaml.print(*args, file=buff)
 		self.assertEqual(b, buff.getvalue())
 
-	def test_str_style_pick(self):
+	def test_str_styles(self):
 		a = pyaml.dump(data_str_multiline)
 		b = pyaml.dump(data_str_multiline, string_val_style='|')
 		self.assertEqual(a, b)
 		b = pyaml.dump(data_str_multiline, string_val_style='plain')
 		self.assertNotEqual(a, b)
+		c = pyaml.dump(data_str_multiline, string_val_style='literal')
+		self.assertNotEqual(c, a)
+		self.assertNotEqual(c, b)
 		self.assertTrue(pyaml.dump('waka waka', string_val_style='|').startswith('|-\n'))
-		self.assertEqual(pyaml.dump(dict(a=1), string_val_style='|'), 'a: 1\n')
+
+		a = pyaml.dump(data_int := dict(a=123))
+		self.assertEqual(a, 'a: 123\n')
+		self.assertEqual(pyaml.dump(data_int, string_val_style='|'), a)
+		self.assertEqual(pyaml.dump(data_int, string_val_style='literal'), a)
+
+		a = pyaml.dump(data_str := dict(a='123'))
+		b = pyaml.dump(data_str, string_val_style='|')
+		self.assertEqual(a, "a: '123'\n")
+		self.assertEqual(self.flatten(data_str), self.flatten(yaml.safe_load(a)))
+		self.assertNotEqual(a, b)
+		self.assertEqual(self.flatten(data_str), self.flatten(yaml.safe_load(b)))
 
 	def test_colons_in_strings(self):
 		val1 = {'foo': ['bar:', 'baz', 'bar:bazzo', 'a: b'], 'foo:': 'yak:'}
@@ -362,6 +367,7 @@ class DumpTests(unittest.TestCase):
 		val1_str = pyaml.dump(val1)
 		val2 = yaml.safe_load(val1_str)
 		self.assertEqual(val1, val2)
+		self.assertIn('key: word1 word2 word3', val1_str)
 
 	def test_empty_strings(self):
 		val1 = {'key': ['', 'stuff', '', 'more'], '': 'value', 'k3': ''}
@@ -394,16 +400,27 @@ class DumpTests(unittest.TestCase):
 		val3 = yaml.safe_load(val2_str)
 
 	def test_namedtuple(self):
-		TestTuple = namedtuple('TestTuple', 'y x z')
+		TestTuple = cs.namedtuple('TestTuple', 'y x z')
 		val = TestTuple(1, 2, 3)
-		val_str = pyaml.dump(val)
-		self.assertEqual(val_str, u'y: 1\nx: 2\nz: 3\n') # namedtuple order was preserved
+		val_str = pyaml.dump(val, sort_keys=False)
+		self.assertEqual(val_str, 'y: 1\nx: 2\nz: 3\n') # namedtuple order was preserved
 
 	def test_ordereddict(self):
-		d = OrderedDict((i, '') for i in reversed(range(10)))
-		lines = pyaml.dump(d).splitlines()
+		d = cs.OrderedDict((i, '') for i in reversed(range(10)))
+		lines = pyaml.dump(d, sort_keys=False).splitlines()
 		self.assertEqual(lines, list(reversed(sorted(lines))))
 
+	def test_enum(self):
+		c = test_const.heartbeat
+		d1 = {'a': c, 'b': c.value, c: 'testx'}
+		self.assertEqual(d1['a'], d1['b'])
+		s = pyaml.dump(d1)
+		d2 = yaml.safe_load(s)
+		self.assertEqual(d1['a'], d2['a'])
+		self.assertEqual(d1['a'], c)
+		self.assertEqual(d1[c], 'testx')
+		self.assertIn('a: 123 # test_const.heartbeat', s)
+
 	def test_pyyaml_params(self):
 		d = {'foo': 'lorem ipsum ' * 30} # 300+ chars
 		for w in 40, 80, 200:
@@ -414,12 +431,13 @@ class DumpTests(unittest.TestCase):
 
 	def test_multiple_docs(self):
 		docs = [yaml.safe_load(large_yaml), dict(a=1, b=2, c=3)]
-		docs_str = pyaml.dump_all(docs, vspacing=[3, 2])
+		docs_str = pyaml.dump_all(docs, vspacing=True)
 		self.assertTrue(docs_str.startswith('---'))
-		self.assertIn('---\n\n\n\na: 1\n\n\n\nb: 2\n\n\n\nc: 3\n', docs_str)
-		docs_str2 = pyaml.dump(docs, vspacing=[3, 2], multiple_docs=True)
+		self.assertIn('---\n\na: 1\n\nb: 2\n\nc: 3\n', docs_str)
+
+		docs_str2 = pyaml.dump(docs, vspacing=True, multiple_docs=True)
 		self.assertEqual(docs_str, docs_str2)
-		docs_str2 = pyaml.dump(docs, vspacing=[3, 2])
+		docs_str2 = pyaml.dump(docs, vspacing=True)
 		self.assertNotEqual(docs_str, docs_str2)
 		docs_str2 = pyaml.dump_all(docs, explicit_start=False)
 		self.assertFalse(docs_str2.startswith('---'))
@@ -429,10 +447,151 @@ class DumpTests(unittest.TestCase):
 
 	def test_ruamel_yaml(self):
 		try: from ruamel.yaml import YAML
-		except ImportError: return unittest.skip('No ruamel.yaml module to test it')
+		except ImportError: self.skipTest('No ruamel.yaml module to test it')
 		data = YAML(typ='safe').load(large_yaml)
 		yaml_str = pyaml.dump(data)
 
+	def test_dump_stream_kws(self):
+		data = [1, 2, 3]
+		buff1, buff2 = io.StringIO(), io.StringIO()
+		pyaml.dump(data, dst=buff1)
+		pyaml.dump(data, stream=buff2)
+		self.assertEqual(buff1.getvalue(), buff2.getvalue())
+
+		buff1.seek(0); buff1.truncate()
+		pyaml.dump(data, dst=buff1, stream=buff1)
+		self.assertEqual(buff1.getvalue(), buff2.getvalue())
+		ys = pyaml.dump(data, dst=str, stream=str)
+		self.assertEqual(ys, buff2.getvalue())
+
+		buff1.seek(0); buff1.truncate(); buff2.seek(0); buff2.truncate()
+		with self.assertRaises(TypeError):
+			pyaml.dump(data, dst=buff1, stream=buff2)
+		with self.assertRaises(TypeError):
+			pyaml.dump(data, dst=str, stream=buff2)
+		self.assertEqual(buff1.getvalue(), '')
+		self.assertEqual(buff2.getvalue(), '')
+
+	def test_vspacing_limits(self):
+		itm = self.yaml_var('''
+			builtIn: 1
+			datasource:
+				type: grafana
+				uid: -- Grafana --
+			enable: yes
+			hide: yes
+			iconColor: rgba(0, 211, 255, 1)
+			name: Annotations & Alerts
+			type: dashboard''')
+		ys = pyaml.dump(dict(mylist=[itm]*10))
+		self.assertEqual(
+			self.empty_line_list(ys),
+			[1, 11, 21, 31, 41, 51, 61, 71, 81, 91] )
+
+		ys = self.yaml_var('''
+			panels:
+				- datasource:
+						type: datasource
+						uid: grafana
+					fieldConfig:''', raw=True)
+		for n in range(60): ys += '\n' + '  '*3 + f'field{n}: value-{n}'
+		ys = pyaml.dump(yaml.safe_load(ys), vspacing=dict(oneline_split=True))
+		self.assertEqual(self.empty_line_list(ys), list(range(4, 126, 2)))
+		ys = pyaml.dump(yaml.safe_load(ys))
+		self.assertEqual(self.empty_line_list(ys), [4])
+
+	def test_split_block_types(self):
+		data = self.yaml_var('''
+			test: # list items should be split
+				- key1: 1
+					key2: A
+				- key1: 1
+					key2: A
+				- key1: 1
+					key2: A
+			test2: # list items should be split
+				- key1:
+						key11: A
+						key12: B
+				- key2:
+						key21: A
+				- val
+			test3: # only top-level keys should be split
+				key1:
+					key11:
+					key12:
+				key2:
+					- key21: A
+					- key22: B
+					- key23: C
+				key3:
+					- key31:
+					- key32:
+					- key33:
+			''')
+		ys = pyaml.dump(data, vspacing=dict(split_lines=0))
+		self.assertEqual(self.empty_line_list(ys), [1, 4, 7, 10, 12, 16, 19, 21, 23, 27, 32])
+
+	def test_anchor_cutoff(self):
+		data = self.yaml_var('''
+			similique-natus-inventore-deserunt-amet-explicabo-cum-accusamus-temporibus:
+				quam-nulla-dolorem-dolore-velit-quis-deserunt-est-ullam-exercitationem:
+					culpa-quia-incidunt-accusantium-ad-dicta-nobis-rerum-veritatis: &test
+						test: 1
+			similique-commodi-aperiam-libero-error-eos-quidem-eius:
+				ipsam-labore-enim,-vero-voluptatem-eaque-dolores-blanditiis-recusandae:
+					quas-atque-maxime-itaque-ullam-sequi-suscipit-quis-vitae-veritatis: *test''')
+		ys = pyaml.dump(data, force_embed=False)
+		for c in '&', r'\*':
+			self.assertTrue(m := re.search(fr'(?<= ){c}\S+', ys))
+			self.assertLess(len(m[0]), 50)
+			self.assertIn('similique', m[0])
+			self.assertIn('veritatis', m[0])
+
+		data = dict(test1=dict(test2=(v := dict(a=1, b=2, c=3))), test3=dict(test4=v))
+		ys = pyaml.dump(data, force_embed=False)
+		for c in '&', r'\*':
+			self.assertTrue(m := re.search(fr'(?<= ){c}\S+', ys))
+			self.assertLess(len(m[0]), 30)
+			self.assertEqual(len(re.findall(r'test\d', m[0])), 2)
+
+	def test_group_oneline_values(self):
+		data = self.yaml_var('''
+			similique-natus: 1
+			similique-commodi:
+				aperiam-libero: 2
+			'vel praesentium quo':
+				exercitationem debitis: porro beatae id
+				rerum commodi ipsum: nesciunt veritatis
+				amet quaerat:
+					assumenda: odio tenetur saepe
+			"111": digit-string
+			deserunt-est-2: asdasd
+			deserunt-est-1: |
+				line1
+				line2
+			culpa-quia: 1234
+			deserunt-est-3: asdasd
+			10: test1
+			200: test
+			30: test2''')
+		# Sort and oneline-group
+		ys1 = pyaml.dump( data,
+			sort_dicts=pyaml.PYAMLSort.oneline_group,
+			vspacing=dict(split_lines=0, split_count=0) )
+		self.assertEqual(self.empty_line_list(ys1), [8, 12, 15, 19])
+		# No oneline-group by default
+		ys2 = pyaml.dump(data, vspacing=dict(split_lines=0, split_count=0))
+		self.assertNotEqual(ys1, ys2)
+		self.assertEqual( self.empty_line_list(ys2),
+			[1, 4, 6, 9, 11, 13, 15, 17, 21, 23, 25, 27, 29] )
+		# No-sort oneline-group overrides oneline-split for conseq oneliners
+		ys3 = pyaml.dump(data, sort_keys=False, vspacing=dict(
+			split_lines=0, split_count=0, oneline_group=True, oneline_split=True ))
+		self.assertNotEqual(ys1, ys3)
+		self.assertNotEqual(ys2, ys3)
+		self.assertEqual(self.empty_line_list(ys3), [1, 4, 8, 11, 14, 18])
+
 
 if __name__ == '__main__':
 	unittest.main()
